#!/usr/bin/env python3
"""Linkulator"""

# If this script contains bugs, blame cmccabe and asdf.

import curses
import curses.textpad as textpad
import getpass
import locale
import signal
from subprocess import call
import sys
import textwrap
from datetime import datetime
from shutil import which
from time import time
from typing import Tuple
from urllib.parse import urlparse

import config
import data


# linkdata columns:
# id (if parent), username, datestamp, parent-id, category, link-url, link-title
LinkData = data.LinkData()
link_data: list = LinkData.link_data


# VIEWS


def view_categories(categories, cols) -> Tuple[str, list[str]]:
    """Produces categories screen display data. Returns as tuple of header and
    content."""

    thead = "{:>4s}  New  {}".format("ID#", "Category")

    content: list[str] = []
    newmarker = "*"
    name_cols = cols - 11

    for i, record in enumerate(categories):
        index_number = i + 1
        name = textwrap.shorten(record["name"], width=name_cols, placeholder="...")
        newmarker = "*" if record["last_updated"] >= config.USER.lastlogin else " "
        content.append(
            "{:4d}   {}   {} ({})".format(
                index_number,
                newmarker,
                name,
                record["count"],
            )
        )

    if not content:
        thead = ""
        content = "\nThere are no posts yet - enter p to post a new link\n".splitlines()
    return thead, content


def view_links(links, category_name, cols) -> Tuple[str, list[str]]:
    """Produces links screen display data. Accepts links list, category name and
    max column width. Returns a tuple of header and content."""

    max_author_cols = max([len(link["link_author"]) for link in links])
    author_cols = max(max_author_cols, 6)

    description_cols = cols - 18 - author_cols - 9 - 1
    # description_cols calulated as: the terminal width, minus the width of ID
    # and Date fields and padding, minus the max name length, minus Resp field
    # and padding width, minus the unread mark width

    thead = "  {}\n {:>3s}  {:>10s}  {:<{author_cols}s}  {:<5}  {:<s}".format(
        category_name.title(),
        "ID#",
        "Date",
        "Author",
        "#Repl",
        "Description",
        author_cols=author_cols,
    )

    content: list[str] = []

    for i, link in enumerate(links):
        description = textwrap.shorten(
            link["description"], width=description_cols, placeholder="..."
        )
        newmarker = "*" if link["has_new_replies"] else ""
        date = datetime.fromtimestamp(float(link["link_timestamp"])).strftime(
            "%Y-%m-%d"
        )

        content.append(
            " {:3d}  {:>10s}  {:<{author_cols}s}  [{:3d}]  {:s}{}".format(
                i + 1,
                date,
                link["link_author"],
                link["reply_count"],
                description,
                newmarker,
                author_cols=author_cols,
            )
        )

    return thead, content


def view_post(post, cols) -> list[str]:
    """Produces post screen display data. Accepts post id and max column
    width. Returns content list."""

    output: list = []

    # set up line wrapping
    line_wrapper = textwrap.TextWrapper(
        initial_indent=" " * 2, subsequent_indent=" " * 21, width=cols
    )

    # post detail view
    output.append("  {:<12}: {}".format("Title", post["title"]))
    output.append("  {:<12}: {}".format("Link", post["url"]))
    output.append("  {:<12}: {}".format("Category", post["category"]))
    output.append("  {:<12}: {}".format("User", post["author"]))
    date = datetime.fromtimestamp(float(post["timestamp"])).strftime("%c")
    output.append("  {:<12}: {}".format("Date", date))

    # post reply view
    if post["replies"]:
        output.append("\n  {}:\n".format("Replies"))
        for replies in post["replies"]:
            reply_author = replies[1]
            reply_date = datetime.fromtimestamp(float(replies[2])).isoformat(
                sep=" ", timespec="minutes"
            )
            reply_text = replies[6]
            output.append(
                line_wrapper.fill(
                    "  {} {}: {}".format(reply_date, reply_author, reply_text)
                )
            )
    else:
        output.append("\n  No replies yet.  Be the first!")
    return output


def view_search_results(keyword: str, search_results: list):
    """Produces search results display data. Accepts search keyword, search
    results and max column width. Returns list of strings."""
    # TODO: update to produce list of strings
    print(
        "\nShowing results for {}\n\n{:>4s}  {:<15s}{:<12s}{:<13s}".format(
            keyword, "ID#", "DATE", "AUTHOR", "DESC"
        )
    )

    for display_index, record in enumerate(search_results, start=1):
        date = datetime.fromtimestamp(float(record[2])).strftime("%Y-%m-%d")
        author = record[1]
        desc = record[6]
        print("{:4d}  {:<15s}{:<12s}{:<13s}".format(display_index, date, author, desc))
    print("")


# CONTROLS


def search():
    """Control for the search function"""
    while True:
        print("")
        keyword = input("Enter your search (or leave empty to cancel): ")
        if keyword == "":
            print("Search cancelled\n")
            return
        search_results = LinkData.search(keyword)
        if not search_results:
            print("No results found\n")
            return
        while True:
            view_search_results(keyword, search_results)
            option = input(
                "Enter a post ID to see its thread, {} to start a new search, {} to go back, or {} to quit: ".format(
                    style_text("s", False, "underline"),
                    style_text("m", False, "underline"),
                    style_text("q", False, "underline"),
                )
            ).lower()
            if option == "q":
                graceful_exit()
            if option == "m":
                return
            if option == "s":
                break
            try:
                if 1 <= int(option) <= len(search_results):
                    menu_view_thread_details(search_results[int(option) - 1][0])
                else:
                    raise IndexError("Invalid post ID")
            except (KeyError, ValueError, IndexError):
                # Catch a Post ID that is not in the thread list or is not a number
                print("{}".format(style_text("Invalid entry", False, "bold")))


def open_link_in_browser(url):
    """Attempts to view the specified URL in the configured browser"""
    if which(config.USER.browser) is None:
        print(
            "Sorry, "
            + config.USER.browser
            + " is not installed on your system.  Ask your sysadmin to install it."
        )
        return

    url_scheme = urlparse(url).scheme
    if url_scheme in ["gopher", "https", "http"]:
        call([config.USER.browser, url])
    else:
        print("Sorry, that url doesn't start with gopher://, http:// or https://")
        try_anyway = input(
            "Do you want to try it in {} anyway? Y/[N]".format(config.USER.browser)
        ).lower()
        if try_anyway == "y":
            call([config.USER.browser, url])


def new_reply(stdscr, post_id):
    # TODO
    pass


# def reply(parent_id):
#    """Prompt for reply, validate input, save validated input to disk and update
#    link_data. Calls view_thread when complete."""
#    while True:
#        comment = input("Enter your comment (or leave empty to abort): ")
#        if comment == "":
#            input("Reply aborted. Hit [Enter] to continue.")
#            break
#        if not is_valid_input(comment):
#            print(
#                "Entries consisting of whitespace, or containing pipes, '|',  are "
#                "not valid.Please try again."
#            )
#        else:
#            record = data.LinkDataRecord(
#                username=getpass.getuser(),
#                timestamp=str(time()),
#                parent_id=parent_id,
#                link_title_or_comment=comment,
#            )
#            LinkData.add(record)
#            input("Reply added.  Hit [Enter] to return to thread.")
#            break


def is_valid_input(entry: str) -> bool:
    """Determine validity of an input string

    >>> is_valid_input("valid")
    True
    >>> is_valid_input("Not|valid")
    False
    >>> is_valid_input("  ")
    False
    """
    if "|" in entry:
        return False
    if entry.strip() == "":
        return False
    return True


def is_correct_category(entry: str) -> bool:
    """Make sure the user purposefully created a new category and not by
    accident (mistyped, tried to use category number instead of name)"""
    if entry not in [record["name"] for record in LinkData.categories]:
        question = "Do you want to create a new category '{}'? Y/[N]".format(entry)
        answer = input(question).lower().strip()
        return answer != "" and answer[0] == "y"
    return True


def get_input(item: str) -> str:
    """Get user input with the specified prompt, validate and return it, or
    break if invalid"""
    while True:
        i: str = input("{}: ".format(style_text(item, True, "underline")))
        if i == "":
            raise ValueError("Empty field")
        if i == "?":
            category_text = (
                "No categories yet, feel free to make a new one!"
                if not LinkData.categories
                else ", ".join(sorted(record["name"] for record in LinkData.categories))
            )
            print(
                "Available categories: {}\n"
                "♻️  Please help keep Linkulator tidy".format(category_text)
            )
            continue
        if is_valid_input(i):
            return i
        print(
            "Entries consisting of whitespace, or containing pipes, '|',  are "
            "not valid.Please try again."
        )


def post_link() -> int:
    """Handles the link posting process"""
    print("\nPost a link by entering the details below.")
    print(
        "Enter {} for a list of categories. Enter an empty field to cancel.\n".format(
            style_text("?", False, "underline")
        )
    )

    try:
        url = get_input("URL")
        category = get_input("Category")
        while not is_correct_category(category):
            category = get_input("Category")
        title = get_input("Title")
    except ValueError:
        print("Post cancelled")
        return -1

    record = data.LinkDataRecord(
        username=getpass.getuser(),
        timestamp=str(time()),
        category=category,
        link_URL=url,
        link_title_or_comment=title,
    )

    return LinkData.add(record)


class Output:
    """Menu content that is output to the curses screen, plus a method to
    calculate how it is displayed"""

    def __init__(self):
        self.thead = None
        self.content = None
        self.length = None
        self.miny = None
        self.maxy = None
        self.scrollminy = 0
        self.viewslice = None

    def reset(self):
        self.scrollminy = 0

    def calc_dimensions(self):
        if not self.content:
            raise ValueError("Can't calculate nonexistant data")
        self.length = len(self.content)
        if self.thead:
            self.miny = 3
        else:
            self.miny = 2
        self.maxy = curses.LINES - 3
        self.viewslice = slice(self.scrollminy, self.scrollminy + self.maxy)


class Menu:
    """Class describing the data contained in each menu level. A level is like a
    menu screen in a hierarchy."""

    def __init__(self, name):
        self.name = name
        self.data = None
        self.output = Output()
        self.selected_key = None


class MenuHierarchy:
    """Class containing all menus, describing their hierarchy and methods to
    handle movement between levels"""

    def __init__(self):
        # primary menu hierarchy: categories -> links -> post
        categories = Menu("categories")
        links = Menu("links")
        post = Menu("post")
        self.primary_hierarchy = [categories, links, post]

        # search menu hierarchy: search_results -> found_post
        search_results = Menu("search_results")
        found_post = Menu("found_post")
        self.search_hierarchy = [search_results, found_post]

        self.is_viewing_primary_hierarchy = True

        self.primary_index = 0
        self.search_index = 0
        self.active = self.primary_hierarchy[self.primary_index]

    def back(self):
        """goes back up the menu hierarchy"""
        if self.is_viewing_primary_hierarchy:
            if self.primary_index > 0:
                self.primary_index -= 1
                self.active = self.primary_hierarchy[self.primary_index]
        else:
            if self.search_index > 0:
                self.search_index -= 1
                self.active = self.search_hierarchy[self.search_index]
            else:
                self.active = self.primary_hierarchy[self.primary_index]

    def forward(self):
        """returns forward through previously visited menu hierarchy"""
        if self.is_viewing_primary_hierarchy:
            if self.primary_index < len(self.primary_hierarchy) - 1:
                if self.primary_hierarchy[self.primary_index + 1].data:
                    self.primary_index += 1
                    self.active = self.primary_hierarchy[self.primary_index]
        else:
            if self.search_index < len(self.search_hierarchy) - 1:
                if self.search_hierarchy[self.search_index + 1].data:
                    self.search_index += 1
                    self.active = self.search_hierarchy[self.search_index]

    def set_active_menu_data(self, LinkData):
        """get active menu data based on menu hierarchy settings"""
        if self.active.name == "categories":
            self.active.data = LinkData.categories
            (
                self.active.output.thead,
                self.active.output.content,
            ) = view_categories(self.active.data, curses.COLS)

        elif self.active.name == "links":
            self.primary_hierarchy[self.primary_index + 1].data = None
            self.active.data = LinkData.get_links_by_category_name(
                self.active.selected_key
            )
            (
                self.active.output.thead,
                self.active.output.content,
            ) = view_links(self.active.data, self.active.selected_key, curses.COLS)

        elif self.active.name == "post":
            self.active.data = LinkData.get_post(self.active.selected_key)
            self.active.output.content = view_post(self.active.data, curses.COLS)

        elif self.active.name == "search_results":
            pass
        elif self.active.name == "search_results_thread_details":
            pass
        else:
            raise ValueError("This shouldn't happen?")


class Status:
    def __init__(self):
        self.message: str
        self.action: str


def term_resize(stdscr):
    """resizes the provided stdscr to the current size"""
    # TODO: ensure resize can decrement bodyscrollmin by the appropriate amount
    # if there is free space available
    stdscr.clear()
    curses.resizeterm(*stdscr.getmaxyx())
    curses.flushinp()
    curses.update_lines_cols()


def menu_system(stdscr):
    """main loop of program, prints data and takes action based on user input"""
    locale.setlocale(locale.LC_ALL, "")
    code = locale.getpreferredencoding()
    curses.use_default_colors()
    curses.curs_set(0)

    title = "Linkulator".encode(code)

    menus = MenuHierarchy()
    menus.set_active_menu_data(LinkData)
    menus.active.output.calc_dimensions()

    status = Status()
    status.message = ""
    status.action = ""

    while True:
        curses.update_lines_cols()
        # TODO: consider stdscr.erase()

        # print title
        stdscr.addstr(0, 1, title)
        stdscr.clrtoeol()

        # TODO: print subtitle?
        # stdscr.addstr(1, 4, "All categories".encode(code))

        # print header
        if menus.active.output.thead:
            stdscr.addstr(
                menus.active.output.miny - 1,
                0,
                menus.active.output.thead.encode(code),
            )
            stdscr.clrtoeol()

        # print body
        menus.active.output.calc_dimensions()

        count = 0
        for i, line in enumerate(
            menus.active.output.content[menus.active.output.viewslice]
        ):
            # guard just in case we try to print beyond the window because of
            # some resize stuff
            if i >= curses.LINES:
                break
            count += count
            stdscr.addstr(i + menus.active.output.miny, 0, line.encode(code))
            stdscr.clrtoeol()

        # this is meant to clear to the second last line, but...
        # just erase to the bottom for now
        stdscr.clrtobot()

        # print status
        if status.message:
            status_text = "{} - {}".format(status.message, status.action)
            stdscr.addstr((curses.LINES - 1), 0, status_text.encode(code))
            stdscr.clrtoeol()
        else:
            stdscr.move(curses.LINES - 1, 0)
            stdscr.clrtoeol()
        # clear message for next loop
        status.message = ""

        # refresh screen output with all changes made
        stdscr.refresh()

        try:
            k = stdscr.getkey()
        except (curses.error):
            status.message = str(curses.error)
            status.action = str(k)
            k = None
            continue

        if k == "KEY_RESIZE":
            term_resize(stdscr)
        elif k in [":", " "]:
            stdscr.addch(curses.LINES - 1, 0, ":")
            stdscr.clrtoeol()
            stdscr.refresh()
            iwin = curses.newwin(1, curses.COLS, curses.LINES - 1, 1)
            itxt = textpad.Textbox(iwin)
            itxt.stripspaces = True
            curses.curs_set(1)
            itxt.edit()
            action = itxt.gather()
            handle_action(menus, action, status)
            curses.curs_set(0)
            del itxt
            del iwin
        else:
            handle_action(menus, k, status)


def handle_action(menus, action, status):
    int_action = None
    try:
        int_action = int(action)
    except ValueError:
        # it's not a number but that's ok
        pass

    if int_action:
        try:
            navigate(menus, int_action)
        except Exception as e:
            status.message = str(e)
    else:
        try:
            do_command(menus, action)
        except Exception as e:
            status.message = str(e)


def navigate(menus, int_action):
    if int_action:
        if menus.active.name == "categories":
            try:
                key = menus.active.data[int_action - 1]["name"]
            except IndexError:
                raise IndexError("Sorry, that category doesn't exist")
                return
        elif menus.active.name == "links":
            try:
                key = menus.active.data[int_action - 1]["post_id"]
            except IndexError:
                raise IndexError("Sorry, that link doesn't exist")
                return
        elif menus.active.name == "search_results":
            # TODO
            pass
        else:
            # no action because it's not a valid menu
            return
        menus.primary_index += 1
        menus.active = menus.primary_hierarchy[menus.primary_index]
        menus.active.selected_key = key
        menus.set_active_menu_data(LinkData)
        menus.active.output.reset()
        menus.active.output.calc_dimensions()


def do_command(menus, action):
    """???"""
    if action in ["j", "KEY_DOWN"]:
        # TODO: put up/down controls in output class?
        if menus.active.output.scrollminy <= (
            menus.active.output.length - menus.active.output.maxy
        ):
            menus.active.output.scrollminy += 1
    elif action in ["k", "KEY_UP"]:
        if menus.active.output.scrollminy > 0:
            menus.active.output.scrollminy -= 1
    elif action in ["b", "back"]:
        menus.back()
    elif action in ["f", "forward"]:
        menus.forward()
    elif action in ["?", "help"]:
        pass
        # print(HELP_TEXT)
    elif action in ["q", "quit", "exit"]:
        graceful_exit()
    elif action in ["c", "create"]:
        # post_id = post_link()
        # if post_id >= 0:
        # TODO: create new post
        # set category_details to the relevant category
        # set the selected index to the index of the new post in category_details
        # set menu level to thread index
        pass
    elif action in ["s", "search"]:
        pass
        # search()
        # TODO: open search screen
        # level.prior_to_search = level.current
        # results are returned to a search list, displayed by setting menu level to search

    elif action in ["r", "reply"] and menus.active.name in [
        "search_result_thread_details",
        "thread_details",
    ]:
        # TODO: need stdscr here
        new_reply(menus.active.selected_key)
    elif action in ["o", "open"] and menus.active.name in [
        "search_result_thread_details",
        "thread_details",
    ]:
        pass
        # open_link_in_browser(thread_details["url"])
        # open link in external program
        # is o the right command?


## GENERAL


def style_text(text: str, is_input: bool, *args) -> str:
    """Style input strings as specified using terminal escape sequences. Returns a styled string"""
    # TODO: not sure if this function will work with curses
    styles = {
        "bold": "\033[1m",
        "dim": "\033[2m",
        "underline": "\033[4m",
        "blink": "\033[5m",  # This is here if you REALLY need it...dont use it.
        # (ctrl+shift+esc+E to enable evil mode.)
        "inverse": "\033[7m",  # Make fg and bg color swap
        "rl_left_fix": "\001" if is_input else "",
        "rl_right_fix": "\002" if is_input else "",
    }
    out = ""
    for arg in args:
        if arg in styles:
            out += styles["rl_left_fix"]
            out += styles[arg]
            out += styles["rl_right_fix"]
    out += text
    out += "{}\033[0m{}".format(styles["rl_left_fix"], styles["rl_right_fix"])
    return out


def graceful_exit():
    """Prints a nice message, performs cleanup and then exits"""
    config.USER.save()
    curses.endwin()
    print("\n\nThank you for linkulating.  Goodbye.\n")
    sys.exit(0)


def signal_handler(sig, frame):
    """handle signals, exiting on SIGINT"""
    graceful_exit()


def main(stdscr):
    """main function - handles argument parsing and calls menu system"""
    signal.signal(signal.SIGINT, signal_handler)
    args = sys.argv[1:]
    config.init()
    if not args:
        menu_system(stdscr)
    elif args[0] in ["-h", "--help", "help"]:
        curses.endwin()
        print(HELP_TEXT)
        sys.exit(0)
    else:
        curses.endwin()
        print("Unknown command: {}".format(args[0]))
        sys.exit(0)
    graceful_exit()


if __name__ == "__main__":
    HELP_TEXT = """
options: -h or --help; or no option to browse links.

Linkulator is a minimalist, commandline link aggregator for small, trusting shell communities.

A few important points about Linkulator:
* Your username is associated with everything you post.  No real anonymity.
* You may ignore other users by adding their username to your ~/.linkulator/ignore file,
  followed by an optional description of why you're ignoring them.
* Link post categories are created dynamically when you submit a link.  Think before you
  create a new category and don't litter the board.
* No files are stored centrally. Each users' contributions are stored in ~/.linkulator/,
  meaning that you may always edit or delete your own files.  Please don't use this ability
  to deceive others.
* Link/reply threads disappear if the original link poster deletes their post; or if you
  put their username in your ignore file.
* Your ~/.linkulator/linkulator.data file must be readable by others, or nobody else will
  see your contributions.
* Linkulator may not work outside of Linux systems.
"""
    curses.wrapper(main)
